# -*- coding:utf-8 -*-
import pytest


class DataProxy(object):
    """
    由于pytest是先导出当前测试文件，再解析配置文件，
    所以如果不使用这个代码懒加载一下factories中的data,
    就会导致ini没有加载就发生了import，即而syspath也没有生效
    """
    @property
    def data(self):
        from factories import data
        return data

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, item):
        return self.data[item]

    def __getattr__(self, item):
        return getattr(self.data, item)


pytestmark = [pytest.mark.item(DataProxy(), a=2, b=1)]


class TestItemPatcher(object):

    pytestmark = [pytest.mark.item(DataProxy(), a=3, b=4)]

    @pytest.mark.item(DataProxy(), a=8, b=9)
    def test_function_scope_mock(self):
        assert DataProxy()["a"] == 8
        assert DataProxy()["b"] == 9

    def test_class_scope_mock(self):
        assert DataProxy()["a"] == 3
        assert DataProxy()["b"] == 4


def test_module_scope_mock():
    assert DataProxy()["a"] == 2
    assert DataProxy()["b"] == 1



# -*- coding:utf-8 -*-
import re
import traceback

from .utils import load


class Attr(object):
    pass


class Prop(object):

    def __init__(self, obj, _name, args, ret_val=None, ret_factory=None, **kwargs):
        """
        :param obj: 被mock的属性拥有者
        :param _name: 被mock的属性名称
        :param ret_val: mock的返回值
        :param ret_factory: mock的返回值生产工厂
        :param asyncable: 被mock的是属性是否是异步的
        :param callable: 被mock的属性是否是可调用的
        :param kwargs: mark传入的kwargs参数
        """
        self.obj = obj
        self.name = _name
        self.ret_val = ret_val
        self.ret_factory = load(ret_factory) \
            if isinstance(ret_factory, str) else ret_factory
        # 优先使用手动指定的标志位
        if "asyncable" in kwargs:
            self.asyncable = kwargs.pop("asyncable")

        if "callable" in kwargs:
            self.callable = kwargs.pop("callable")

        self.args = args or tuple()
        self.kwargs = kwargs or dict()

    def __iter__(self):
        try:
            yield self.obj
            yield self.name
            if self.callable:
                yield self
            else:
                if self.ret_factory:
                    self.ret_val = self.ret_factory()
                yield self.ret_val
        except Exception as e:
            traceback.print_exc()
            raise e

    def __call__(self, *args, **kwargs):
        kwargs.update(self.kwargs)
        if self.ret_factory:
            ret_val = self.ret_factory(*(args + self.args), **kwargs)
        else:
            ret_val = self.ret_val

        if self.asyncable:
            from .compact import get_coroutine
            return get_coroutine(ret_val)
        else:
            return ret_val

    def __str__(self):
        kwargs = self.__dict__.copy()
        # 有可能还没asyncable, callable这两个属性，因为这两个属性是在process_mark时赋予的。
        kwargs.setdefault("asyncable", False)
        kwargs.setdefault("callable", False)
        return "<Prop(name={name} obj={obj} asyncable={asyncable} " \
               "callable={callable} kwargs={kwargs} ret_val={ret_val} " \
               "ret_factory={ret_factory})>".format(**kwargs)

    __repr__ = __str__


def parse(mock, args, kwargs=None, regex=re.compile(r" no[\w\s]+?(['\"])(\w+)\1")):
    """
    将mark转换成prop
    :param mock:
    :param args:
    :param kwargs:
    :param regex: 祖先不存在时用来mock祖先，使用一个正则来将祖先的名称找出
    :return:
    """

    obj_name, _, prop = mock.rpartition(".")

    while True:
        try:
            obj = load(obj_name)
            yield Prop(obj, prop, args, **kwargs)
            break
        except AttributeError as ex:
            ancestor = regex.search(str(ex)).group(2)
            _obj_name, last = re.split(ancestor, obj_name, 1)

            for mock_prop in parse(
                    _obj_name + ancestor, tuple(), {"ret_val": Attr()}):
                yield mock_prop



# -*- coding:utf-8 -*-
data = dict(a=5, b=8)


mock_get_data_session = lambda a, **kwargs: a*2


class TestClass(object):

    @classmethod
    def get_data_session(cls):
        return a

    @classmethod
    def get_data_module(cls):
        return a

    @classmethod
    def get_data_class(cls):
        return a

    @classmethod
    def get_data_function(cls):
        return a


# -*- coding:utf-8 -*-
import pytest

from os import getcwd
from os.path import dirname

pytestmark = [pytest.mark.path(dirname(dirname((dirname(__file__)))))]


@pytest.mark.usefixtures("mock")
def test_root_path():
    assert getcwd() == dirname(dirname(dirname(__file__)))


@pytest.mark.usefixtures("mock")
class TestPathPatcher(object):
    pytestmark = [pytest.mark.path(dirname(dirname(__file__)))]

    def test_current_class_path(self):
        assert getcwd() == dirname(dirname(__file__))

    @pytest.mark.path(dirname(__file__))
    def test_current_path(self):
        assert getcwd() == dirname(__file__)




# -*- coding:utf-8 -*-
import os
import time
import pytest

from functools import partial

from .utils import create_server
from .patcher import build, process


@pytest.fixture(scope="session")
def join_root_dir(pytestconfig):
    return partial(os.path.join, os.path.abspath(
        pytestconfig.getoption("rootdir") or "."))


@pytest.fixture(scope="module")
def server(request):
    old_path = os.getcwd()
    loop = None
    server = None
    try:
        path = os.path.dirname(request.module.__file__)
        loop, server = create_server(path)
        yield server
    finally:
        os.chdir(old_path)
        if server:
            server.server.close()
            server.should_exit = True
            # 等待tick退出
            time.sleep(1)
        if loop:
            loop.stop()


session_mock = build("session", "pytestconfig")
module_mock = build("module")
class_mock = build("class")


@pytest.fixture
def mock(request, session_mock, module_mock, class_mock):
    """封装monkey patch实现mock env和prop"""
    # 这里不能连着写，因为连着写生成器会被马上回收，
    # 回收后会马上调用上下文对象的__exit__，导致触发monkey.undo
    gen = process(request, request)
    yield next(gen)




# -*- coding:utf-8 -*-
from pyaop import Proxy, AOP
from _pytest.mark import MarkDecorator, Mark


__version__ = "0.1.10"


def wrapper(mock_obj_prefix, mock_factory_prefix):

    def call(proxy, prop_name, *args, **kwargs):
        ret_factory = kwargs.pop("ret_factory", None)
        if mock_obj_prefix:
            pn = "{}.{}".format(mock_obj_prefix, prop_name)
        else:
            pn = prop_name

        if ret_factory and isinstance(ret_factory, str) and mock_factory_prefix:
            fn = "{}.{}".format(mock_factory_prefix, ret_factory)
        else:
            fn = ret_factory
        kwargs["ret_factory"] = fn
        return (pn, ) + args, kwargs
    return call


def prop_alias(mock_obj_prefix=None, mock_factory_prefix="factories"):
    """
    prop mark装饰器别名，指定要mock对象及工厂对象名字的前缀，减少装饰器长度。
    :param mock_obj_prefix:
    :param mock_factory_prefix:
    :return:
    """
    return Proxy(MarkDecorator(Mark("prop", (), {})),
                 before=[AOP.Hook(
                     wrapper(mock_obj_prefix, mock_factory_prefix),
                     ["__call__"])])



# -*- coding:utf-8 -*-
import re
import os
import six
import pytest
import inspect

from functools import partial
from abc import ABCMeta, abstractmethod

from _pytest.mark import Mark
from _pytest.monkeypatch import MonkeyPatch

from .parser import parse
from .utils import load, cache_classproperty, MarkerWrapper, guess, find_children

namespace = {"function": 10, "class": 20, "package": 30, "module": 40, "session": 50}


@six.add_metaclass(ABCMeta)
class Patcher(object):

    @property
    @abstractmethod
    def name(self):
        """
        定义一个名字
        :return:
        """

    @cache_classproperty
    def total_markers(self):
        """
        由于iter_markers会返回所有scope下的mark，
        所以使用这个属性来保证每个markers只处理一次，
        :return:
        :rtype: dict
        """
        return dict()

    def __init__(self, markers):
        self.monkey_patch = MonkeyPatch()
        self.markers = markers

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.monkey_patch.undo()

    def process(self, request):
        fixture_kwargs = dict()
        # 只有PropPatcher才需要注入fixture。
        if self.name == "prop":
            for name, fixture in request._fixture_defs.items():
                # 当前请求作用域适用于该fixture的作用域
                if namespace[request.scope] <= namespace[fixture.scope]:
                    fixture_kwargs[name] = request.getfixturevalue(name)

        for mark in self.markers:
            mark_wrapper = MarkerWrapper(mark)
            # 由于当前request可获得的markers包括大于等于当前作用域的所有marker。
            # 而请求加载顺序的作用域从大到小，如到function作用域时，
            # 其它作用域已经处理过一些mark，这些mark需要过滤掉，
            # 但在执行时若使用了多输入的fixture，会出现请求重复发起的情况，
            # 此时需要重复执行mock， 因此，不能对function级别的mark进行过滤。
            if mark_wrapper not in self.total_markers or \
                    self.total_markers[mark_wrapper] == "function":
                self.process_mark(mark, fixture_kwargs)
                self.total_markers[mark_wrapper] = request.scope

    @abstractmethod
    def process_mark(self, mark, fixture_kwargs):
        """
        个性化定制mark处理方法
        :return:
        """

    @classmethod
    @abstractmethod
    def config_parse(cls, mark):
        """
        config解析方式
        :param mark:
        :return: Mark
        """

    @classmethod
    def from_pytestconfig(cls, pytestconfig):
        mocks = pytestconfig.inicfg.get(cls.name)
        ms = list()
        if mocks:
            for m in mocks.strip().split("\n"):
                if m.strip():
                    mark = cls.config_parse(m.strip())
                    if mark:
                        ms.append(mark)
        return cls(ms)

    @classmethod
    def from_request(cls, request, *args, **kwargs):
        return cls(request.node.iter_markers(cls.name))


class PropPatcher(Patcher):
    """
    用来monkey patch 属性
    """
    name = "prop"
    order = 5

    def guess_attr(self, prop, old, mock, func):
        # 证明old是prop
        if hasattr(old, prop):
            setattr(mock, prop, getattr(old, prop))
        # 如果手动指定了，则使用手动指定的
        elif hasattr(mock, prop):
            pass
        # 否则，自行判断
        else:
            setattr(mock, prop, func(old))

    def process_mark(self, mark, fixture_kwargs):
        kwargs = mark.kwargs.copy()
        # 是否往factory中注入fixture
        fixture_inject = kwargs.pop("fixture_inject", False)
        for mock in parse(mark.args[0], mark.args[1:], kwargs=kwargs):
            if mock.ret_factory and fixture_kwargs and fixture_inject:
                mock.ret_factory = partial(mock.ret_factory, **fixture_kwargs)

            try:
                old = getattr(mock.obj, mock.name, None)
                try:
                    from asyncio import iscoroutinefunction
                    self.guess_attr("asyncable", old, mock, iscoroutinefunction)
                except ImportError:
                    mock.asyncable = False
                self.guess_attr("callable", old, mock, callable)
                # apistellar的依赖注入需要return 的signature
                if mock.callable:
                    if getattr(old, "__annotations__", None):
                        if "return" in old.__annotations__:
                            mock.__signature__ = inspect.Signature(
                                return_annotation=old.__annotations__["return"])
                            mock.__annotations__ = {
                                "return": old.__annotations__["return"]}
            except AttributeError:
                raise RuntimeError("{} has not attr: {}".format(
                    mock.obj, mock.name))
            self.monkey_patch.setattr(*mock, raising=False)

    @classmethod
    def config_parse(cls, mark):
        if "->" in mark:
            mark, ret_factory = mark.split("->", 1)
            kwargs = {"ret_factory": ret_factory}
        else:
            mark, ret_val = mark.split("=", 1)
            kwargs = {"ret_val": guess(ret_val)}

        return Mark(cls.name, tuple([mark]), kwargs)


class ItemPatcher(Patcher):
    """
    用来monkey patch Item
    """
    name = "item"
    order = 4
    mark_config_regex = re.compile(r"(.+?)\[(.+?)\]\s*=\s*?(.+)")

    def process_mark(self, mark, fixture_kwargs):
        for key, val in mark.kwargs.items():
            item = mark.args[0]
            if isinstance(item, str):
                item = load(item)
            self.monkey_patch.setitem(item, key, val)

    @classmethod
    def config_parse(cls, mark_str):
        mth = cls.mark_config_regex.search(mark_str)
        if mth:
            prop_name, key, val = mth.groups()
            return Mark(cls.name, tuple([prop_name]), {eval(key): guess(val)})


class EnvPatcher(Patcher):
    """
    用来monkey patch 环境变量
    """
    name = "env"
    order = 3

    def process_mark(self, mark, fixture_kwargs):
        prepend = mark.kwargs.pop("prepend", None)
        for key, val in mark.kwargs.items():
            self.monkey_patch.setenv(key, val, prepend)

    @classmethod
    def config_parse(cls, mark_str):
        key, val = mark_str.split("=", 1)
        return Mark(cls.name, tuple(), {key: guess(val)})


class PathPatcher(Patcher):
    """
        用来monkey patch 目录
        """
    name = "path"
    order = 1

    def process_mark(self, mark, fixture_kwargs):
        self.monkey_patch.chdir(mark.args[0])

    @classmethod
    def config_parse(cls, mark_str):
        return Mark(cls.name, tuple([guess(mark_str)]), dict())


class SysPathPatcher(PathPatcher):
    """
    用来monkey patch sys path
    """
    name = "syspath"
    order = 2

    def process_mark(self, mark, fixture_kwargs):
        self.monkey_patch.syspath_prepend(os.path.abspath(mark.args[0]))


def process(fromobj, request=None, load_from="request",
            patchers=sorted(find_children(Patcher), key=lambda x: x.order)):
    if not patchers:
        yield

    patcher = patchers[0]
    with getattr(patcher, "from_%s" % load_from)(fromobj) as patcher:
        patcher.process(request)
        # 使用一个变量来指向生成器，防止在整个函数返回之前被gc
        gen = process(fromobj, request, load_from, patchers[1:])
        yield next(gen)


def build(scope, load_from="request"):
    def mock(request, pytestconfig):
        gen = process(locals()[load_from], request=request, load_from=load_from)
        yield next(gen)

    return pytest.fixture(scope=scope, name="%s_mock" % scope)(mock)


import pytest


class TestParse(object):

    @pytest.mark.prop("os.b.c.d", ret_val=1)
    def test_cascade_mock(self):
        import os
        assert hasattr(os, "b")
        assert hasattr(os.b, "c")
        assert os.b.c.d == 1



import pytest
from pytest_apistellar.utils import load


class TestLoad(object):

    def test_load_nomal(self):
        assert load("os.path").__name__ == "posixpath"

    def test_load_not_found_attr(self):
        with pytest.raises(AttributeError):
            load("os.path.a")

    def test_load_not_found_module(self):
        with pytest.raises(ImportError):
            load("fdsff")


# -*- coding:utf-8 -*-
import re
import time
import socket
import warnings
import threading

from functools import wraps, reduce


def run_server(path, container, port=None):
    """
    创建一个简单的server用来测试
    :param path:
    :param port:
    :return:
    """
    try:
        import asyncio
        from apistellar import Application
        from uvicorn.main import Server, HttpToolsProtocol
    except ImportError:
        warnings.warn("Python3.6: apistellar required. ")
        raise
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    app = Application("test", current_dir=path)
    port = port or free_port()
    server = Server(app, "127.0.0.1", port, loop, None, HttpToolsProtocol)
    loop.run_until_complete(server.create_server())

    if server.server is not None:
        container.append(loop)
        container.append(server)
        loop.create_task(server.tick())
        loop.run_forever()


def create_server(path):
    container = []
    th = threading.Thread(target=run_server, args=(path, container))
    th.setDaemon(True)
    th.start()
    # 等待10秒
    for i in range(100):
        time.sleep(0.1)
        if len(container) == 2:
            break
    else:
        raise RuntimeError("子线程启动超时！")
    return container


def free_port():
    """
    Determines a free port using sockets.
    """
    free_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    free_socket.bind(('0.0.0.0', 0))
    free_socket.listen(5)
    port = free_socket.getsockname()[1]
    free_socket.close()
    return port


def load(prop_str):
    """
    返回字符串表示的模块、函数、类、若类的属性等
    :param prop_str: module1.class.function....
    :return: function
    """
    attr_list = []
    # 每次循环将prop_str当模块路径查找，成功则返回，
    # 失败则将模块路径回退一级，将回退的部分转换成属性
    # 至到加载模块成功后依次从模块中提取属性。
    ex = None
    while prop_str:
        try:
            obj = __import__(prop_str, fromlist=prop_str.split(".")[-1])
            for attr in attr_list:
                obj = getattr(obj, attr)
            return obj
        except (AttributeError, ImportError) as e:
            prop_str, _sep, attr_str = prop_str.rpartition('.')
            attr_list.insert(0, attr_str)
            ex = e
    else:
        raise ex


def guess(val):
    """
    通过字符串表达式去猜测要返回的值
    @param val:
    @return:
    """
    try:
        return eval(val)
    except NameError:
        # 如果报错了，可能是字符串描述的模块没有导入，则导入模块
        try:
            # 寻找[或(，来判断是否需要执行方法或函数
            mth = re.search(r"[\[\(]", val)
            # 如果存在，则证明需要调用函数或类
            if mth:
                prop_str, args_str = val[:mth.start()], val[mth.start():]
                prop = load(prop_str)
                prop_name = "_local_prop_name"
                locals()[prop_name] = prop
                return eval(prop_name + args_str)
            else:
                return load(val)
        except (ImportError, NameError, AttributeError):
            return val
    except SyntaxError:
        return val


def cache_property(func):
    """
    缓存属性，只计算一次
    :param func:
    :return:
    """
    @property
    @wraps(func)
    def wrapper(*args, **kwargs):
        if func.__name__ not in args[0].__dict__:
            args[0].__dict__[func.__name__] = func(*args, **kwargs)
        return args[0].__dict__[func.__name__]
    return wrapper


class classproperty(object):
    """
    property只能用于实例方法到实例属性的转换，使用classproperty来支持类方法到类属性的转换
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        return self.func(owner)


def cache_classproperty(func):
    """
    缓存类属性，只计算一次
    :param func:
    :return:
    """
    @classproperty
    @wraps(func)
    def wrapper(*args, **kwargs):
        prop_name = "_" + func.__name__
        if prop_name not in args[0].__dict__:
            setattr(args[0], prop_name, func(*args, **kwargs))
        return args[0].__dict__[prop_name]
    return wrapper


class MarkerWrapper(object):
    """
    Mark类__eq__使用(name, args, kwargs)是否相同来判断，
    无法满足去重的要求，所以使用这个类来包装一下使用id来去重。
    """
    def __init__(self, marker):
        self.marker = marker

    def __hash__(self):
        return id(self.marker)

    def __eq__(self, other):
        if not hasattr(other, "marker"):
            return False

        return id(self.marker) == id(other.marker)

    def __repr__(self):
        return repr(self.marker)

    __str__ = __repr__


def find_children(cls):
    """
    获取所有(component)的子类或其实例。
    :param cls: 父类
    :return:
    """
    return set(reduce(lambda x, y: x.union(set([y])).union(find_children(y)),
                      cls.__subclasses__(), set()))
